import collections
import functools
import itertools
try:
    from urllib import parse
except ImportError:
    import urlparse as parse

from marnadi import Route, descriptors, Header
from marnadi.errors import HttpError
from marnadi.handlers import Handler
from marnadi.utils import cached_property


class Request(collections.Mapping):
    """WSGI request.

    Args:
        environ (dict): PEP-3333 WSGI environ dict.
    """

    if hasattr(collections.Mapping, '__slots__'):
        __slots__ = '_environ', '__weakref__'

    __hash__ = object.__hash__

    __eq__ = object.__eq__

    __ne__ = object.__ne__

    def __init__(self, environ):
        self._environ = environ

    def __getitem__(self, key):
        return self._environ[key]

    def __iter__(self):
        return iter(self._environ)

    def __len__(self):
        return len(self._environ)

    @property
    def input(self):
        return self['wsgi.input']

    @property
    def method(self):
        return self['REQUEST_METHOD']

    @property
    def path(self):
        return self['PATH_INFO']

    @cached_property
    def content_type(self):
        try:
            parts = iter(self['CONTENT_TYPE'].split(';'))
            return Header(next(parts).strip(), **dict(
                map(str.strip, option.split('='))
                for option in parts
            ))
        except KeyError:
            raise AttributeError("content_type is not provided")

    @property
    def content_length(self):
        try:
            return int(self['CONTENT_LENGTH'])
        except KeyError:
            raise AttributeError("content_length is not provided")

    @cached_property
    def headers(self):
        return {
            name.title().replace('_', '-'): value
            for name, value in
            itertools.chain(
                (
                    (env_key, self[env_key])
                    for env_key in ('CONTENT_TYPE', 'CONTENT_LENGTH')
                    if env_key in self
                ),
                (
                    (env_key[5:], env_value)
                    for env_key, env_value in self.items()
                    if env_key.startswith('HTTP_')
                ),
            )
        }

    @cached_property
    def query(self):
        try:
            return dict(parse.parse_qsl(
                self['QUERY_STRING'],
                keep_blank_values=True,
            ))
        except KeyError:
            return {}

    data = descriptors.Data(
        (
            'application/json',
            'marnadi.descriptors.data.decoders.application.json.Decoder',
        ),
        (
            'application/x-www-form-urlencoded',
            'marnadi.descriptors.data.decoders' +
            '.application.x_www_form_urlencoded.decoder',
        ),
    )


class App(object):
    """WSGI application class.

    Instance of this class used as entry point for WSGI requests. Using
    provided routes list it can determine which handler should be called.

    Args:
        routes (iterable): list of :class:`Route`.
    """

    __slots__ = 'routes', 'routes_map'

    def __init__(self, routes=()):
        self.routes_map = {}
        self.routes = self.compile_routes(routes)

    def __call__(self, environ, start_response):
        try:
            request = self.make_request_object(environ)
            handler = self.get_handler(request.path)
            response = handler.send((self, request))
        except HttpError as error:
            response = error
        start_response(
            response.status,
            list(response.headers.items(stringify=True))
        )
        return response

    @staticmethod
    def make_request_object(environ):
        return Request(environ)

    def compile_routes(self, routes, parents=()):
        callback = functools.partial(self.compile_route, parents=parents)
        return list(map(callback, routes))

    def compile_route(self, route, parents=()):
        assert isinstance(route, Route)
        parents = parents + (route, )
        if route.name:
            self.routes_map[route.name] = parents
        if isinstance(route.handler, Handler):
            return route
        try:
            route.handler = self.compile_routes(route.handler, parents=parents)
        except TypeError:
            raise TypeError(
                "Route's handler must be either subclass of Handler " +
                "or sequence of nested subroutes")
        return route

    def route(self, path, name=None, params=None, patterns=None):
        def _decorator(handler):
            route = Route(
                path, handler, name=name, params=params, patterns=patterns)
            self.routes.append(self.compile_route(route))
            return handler
        return _decorator

    def make_path(self, *route_name, **params):
        if len(route_name) != 1:
            raise TypeError(
                "either route_name isn't provided or it isn't a single value")
        return ''.join(
            route.restore_path(**params)
            for route in self.routes_map[route_name[0]]
        )

    def get_handler(self, path, routes=None, params=None):
        """Return handler according to the given path.

        Note:
            If you wish for example automatically redirect all requests
            without trailing slash in URL to URL with persisting one you may
            override this method by raising `HttpError` with 301 status and
            necessary 'Location' header when needed.
        """
        routes = routes or self.routes
        params = params or {}
        for route in routes:
            match = route.match(path)
            if not match:
                continue
            rest_path, url_params = match
            if isinstance(route.handler, list):
                try:
                    return self.get_handler(
                        rest_path,
                        routes=route.handler,
                        params=self._merge_dicts(
                            params.copy(), route.params, url_params),
                    )
                except HttpError:
                    continue
            if not rest_path:
                return route.handler.start(**self._merge_dicts(
                    params, route.params, url_params))
        raise HttpError('404 Not Found')  # matching route not found

    @staticmethod
    def _merge_dicts(target, *sources):
        for source in sources:
            target.update(source)
        return target



import importlib
import weakref

try:
    unicode_str = unicode
except NameError:
    unicode_str = str


def metaclass(mcs):
    def _decorator(cls):
        attrs = dict(vars(cls))
        try:
            if isinstance(cls.__slots__, str):
                slots = (cls.__slots__, )
            else:
                slots = cls.__slots__
            for slot in slots:
                del attrs[slot]
        except AttributeError:
            pass
        for prop in '__weakref__', '__dict__':
            attrs.pop(prop, None)
        return mcs(cls.__name__, cls.__bases__, attrs)
    return _decorator


class CachedDescriptor(object):

    __slots__ = 'cache',

    def __init__(self):
        self.cache = weakref.WeakKeyDictionary()

    def __get__(self, instance, instance_type=None):
        if instance is None:
            return self  # static access
        try:
            return self.cache[instance]
        except KeyError:
            value = self.cache[instance] = self.get_value(instance)
            return value

    def get_value(self, instance):
        raise NotImplementedError


class cached_property(CachedDescriptor):

    __slots__ = 'get', 'set', 'delete', '__doc__'

    def __init__(self, fget=None, fset=None, fdel=None, doc=None):
        super(cached_property, self).__init__()
        self.get = fget
        self.set = fset
        self.delete = fdel
        self.__doc__ = doc

    def get_value(self, instance):
        if self.get is None:
            raise AttributeError("unreadable attribute")
        return self.get(instance)

    def __set__(self, instance, value):
        if self.set is None:
            self.cache[instance] = value
        else:
            self.set(instance, value)
            self.cache.pop(instance, None)

    def __delete__(self, instance):
        if self.delete is not None:
            self.delete(instance)
        self.cache.pop(instance, None)

    def getter(self, getter):
        self.get = getter
        return self

    def setter(self, setter):
        self.set = setter
        return self

    def deleter(self, deleter):
        self.delete = deleter
        return self


class LazyType(type):

    def __call__(cls, path):
        if isinstance(path, cls) or not isinstance(path, str):
            return path
        return super(LazyType, cls).__call__(path)


@metaclass(LazyType)
class Lazy(object):

    __slots__ = '_path', '__weakref__'

    def __init__(self, path):
        self._path = path

    def __call__(self, *args, **kwargs):
        return self._obj(*args, **kwargs)

    def __iter__(self):
        return iter(self._obj)

    def __len__(self):
        return len(self._obj)

    def __str__(self):
        return str(self._obj)

    def __unicode__(self):
        return unicode(self._obj)

    def __bytes__(self):
        return bytes(self._obj)

    def __getitem__(self, item):
        return self._obj[item]

    def __getattr__(self, attr):
        return getattr(self._obj, attr)

    @property
    def __class__(self):
        return self._obj.__class__

    @cached_property
    def _obj(self):
        try:
            module_name, obj_name = self._path.rsplit('.', 1)
        except ValueError:
            module_name, obj_name = self._path, None
        module = importlib.import_module(module_name)
        if obj_name is not None:
            return getattr(module, obj_name)
        return module


def to_bytes(obj, encoding='utf-8', error_callback=None):
    try:
        if isinstance(obj, (bytes, bytearray, memoryview)):
            return bytes(obj)
        if obj is None:
            return b''
        try:
            return obj.__bytes__()
        except AttributeError:
            return unicode_str(obj).encode(encoding=encoding)
    except Exception as error:
        if error_callback is not None:
            error_callback(error)
        raise


def coroutine(fn):
    def _fn(*args, **kwargs):
        co = fn(*args, **kwargs)
        co.send(None)
        return co
    return _fn



from marnadi.utils import metaclass


class DecoderType(type):

    def __call__(cls, request):
        decoder = super(DecoderType, cls).__call__()
        return decoder(request)


@metaclass(DecoderType)
class Decoder(object):

    __slots__ = ()

    def __call__(self, request):
        return b''.join(request.input)



import collections
import itertools

from marnadi.utils import cached_property, CachedDescriptor


class HeadersMixin(collections.Mapping):

    if hasattr(collections.Mapping, '__slots__'):
        __slots__ = '__weakref__',

    def __getitem__(self, header):
        return self._headers[header.title()]

    def __len__(self):
        return len(self._headers)

    def __iter__(self):
        return iter(self._headers)

    __hash__ = object.__hash__

    __eq__ = object.__eq__

    __ne__ = object.__ne__

    @cached_property
    def _headers(self):
        raise ValueError("This property must be set before using")

    def items(self, stringify=False):
        for header, values in self._headers.items():
            for value in values:
                yield header, str(value) if stringify else value

    def values(self, stringify=False):
        for values in self._headers.values():
            for value in values:
                yield str(value) if stringify else value


class ResponseHeaders(HeadersMixin, collections.MutableMapping):

    __slots__ = ()

    def __init__(self, default_headers):
        self._headers = default_headers

    def __delitem__(self, header):
        del self._headers[header.title()]

    def __setitem__(self, header, value):
        self._headers[header.title()] = [value]

    def append(self, header, value):
        self._headers[header.title()].append(value)

    def extend(self, *headers):
        for header in headers:
            self.append(*header)

    def setdefault(self, header, default=None):
        return self._headers.setdefault(header.title(), [default])

    def clear(self, *headers):
        if headers:
            for header in headers:
                try:
                    del self[header]
                except KeyError:
                    pass
        else:
            self._headers.clear()


class Headers(CachedDescriptor, HeadersMixin):

    __slots__ = ()

    def __init__(self, *default_headers, **kw_default_headers):
        super(Headers, self).__init__()
        self._headers = collections.defaultdict(list)
        for header, value in itertools.chain(
            default_headers,
            kw_default_headers.items(),
        ):
            self._headers[header.title()].append(value)

    def get_value(self, instance):
        return ResponseHeaders(default_headers=self._headers.copy())



try:
    from urllib import parse
except ImportError:
    import urlparse as parse


def decoder(request):
    return dict(parse.parse_qsl(
        b''.join(request.input),
        keep_blank_values=True,
    ))



from distutils.core import setup

with open('README.rst') as description:
    long_description = description.read()

setup(
    name='Marnadi',
    version='0.2.2',
    author='Rinat Khabibiev',
    author_email='srenskiy@gmail.com',
    packages=[
        'marnadi',
        'marnadi.descriptors',
        'marnadi.descriptors.data',
        'marnadi.descriptors.data.decoders',
        'marnadi.descriptors.data.decoders.application',
    ],
    url='https://github.com/renskiy/marnadi',
    license='MIT',
    description='Yet another WSGI web framework',
    keywords='WSGI, HTTP, REST',
    long_description=long_description,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Environment :: Web Environment',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)